import json
import uuid
from urllib.parse import urlparse
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.contrib.auth import get_user_model
from django.db import models
from django.db.models import BinaryField, DateField
from django.db.models.base import ModelBase
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.urls import reverse_lazy, get_resolver
from django.utils.datastructures import MultiValueDictKeyError
from django.utils.decorators import classonlymethod

from djangoldp.fields import LDPUrlField
from djangoldp.permissions import LDPPermissions
import logging


logger = logging.getLogger('djangoldp')


class LDPModelManager(models.Manager):
    # an alternative to all() which exlcudes external resources
    def local(self):
        queryset = super(LDPModelManager, self).all()
        internal_ids = [x.pk for x in queryset if not Model.is_external(x)]
        return queryset.filter(pk__in=internal_ids)


class Model(models.Model):
    urlid = LDPUrlField(blank=True, null=True, unique=True)
    is_backlink = models.BooleanField(default=False,
                                      help_text='(DEPRECIATED) set automatically to indicate the Model is a backlink')
    allow_create_backlink = models.BooleanField(default=True,
                                                help_text='set to False to disable backlink creation after Model save')
    objects = LDPModelManager()

    def __init__(self, *args, **kwargs):
        super(Model, self).__init__(*args, **kwargs)

    @classmethod
    def get_view_set(cls):
        '''returns the view_set defined in the model Meta or the LDPViewSet class'''
        view_set = getattr(cls._meta, 'view_set', getattr(cls.Meta, 'view_set', None))
        if view_set is None:
            from djangoldp.views import LDPViewSet
            view_set = LDPViewSet
        return view_set

    @classmethod
    def get_container_path(cls):
        '''returns the url path which is used to access actions on this model (e.g. /users/)'''
        path = getattr(cls._meta, 'container_path', getattr(cls.Meta, 'container_path', None))
        if path is None:
            path = "{}s".format(cls._meta.object_name.lower())

        return cls.__clean_path(path)

    def get_absolute_url(self):
        return Model.absolute_url(self)

    @classonlymethod
    def absolute_url(cls, instance_or_model):
        if isinstance(instance_or_model, ModelBase) or instance_or_model.urlid is None or instance_or_model.urlid == '':
            return '{}{}'.format(settings.SITE_URL, Model.resource(instance_or_model))
        else:
            return instance_or_model.urlid

    def get_container_id(self):
        return Model.container_id(self)

    @classonlymethod
    def resource(cls, instance_or_model):
        if isinstance(instance_or_model, ModelBase):
            return cls.container_id(instance_or_model)
        else:
            return cls.resource_id(instance_or_model)

    @classonlymethod
    def resource_id(cls, instance):
        r_id = "{}{}".format(cls.container_id(instance), getattr(instance, cls.slug_field(instance), ""))
        return cls.__clean_path(r_id)

    @classonlymethod
    def slug_field(cls, instance_or_model):
        if isinstance(instance_or_model, ModelBase):
            object_name = instance_or_model.__name__.lower()
        else:
            object_name = instance_or_model._meta.object_name.lower()
        view_name = '{}-detail'.format(object_name)
        try:
            slug_field = '/{}'.format(get_resolver().reverse_dict[view_name][0][0][1][0])
        except MultiValueDictKeyError:
            slug_field = Model.get_meta(instance_or_model, 'lookup_field', 'pk')
        if slug_field.startswith('/'):
            slug_field = slug_field[1:]
        return slug_field

    @classonlymethod
    def container_id(cls, instance):
        if isinstance(instance, cls):
            path = instance.get_container_path()
        else:
            view_name = '{}-list'.format(instance._meta.object_name.lower())
            path = get_resolver().reverse(view_name)

        path = cls.__clean_path(path)

        return path

    class Meta:
        default_permissions = ('add', 'change', 'delete', 'view', 'control')
        abstract = True
        depth = 0

    @classonlymethod
    def resolve_id(cls, id):
        '''
        Resolves the id of a given path (e.g. /container/1/)
        Raises Resolver404 if the path cannot be found, ValidationError if the path is for a model base
        and an ObjectDoesNotExist exception if the resource does not exist
        '''
        id = cls.__clean_path(id)
        match = get_resolver().resolve(id)
        kwargs = match.kwargs
        view = match.func

        if match.url_name.endswith('-list') or len(match.kwargs.keys()) == 0:
            raise ValidationError('resolve_id received a path for a container or nested container')
        return view.initkwargs['model'].objects.get(**kwargs)

    @classonlymethod
    def resolve_parent(cls, path):
        split = path.strip('/').split('/')
        parent_path = "/".join(split[0:len(split) - 1])
        return Model.resolve_id(parent_path)

    @classonlymethod
    def resolve_container(cls, path):
        '''retruns the model container of passed URL path'''
        path = cls.__clean_path(path)
        view, args, kwargs = get_resolver().resolve(path)
        return view.initkwargs['model']

    @classonlymethod
    def resolve(cls, path):
        '''
        resolves the containing model and associated id in the path. If there is no id in the path returns None
        :param path: a URL path to check
        :return: the container model and resolved id in a tuple
        '''
        if settings.BASE_URL in path:
            path = path[len(settings.BASE_URL):]
        container = cls.resolve_container(path)
        try:
            resolve_id = cls.resolve_id(path)
        except:
            resolve_id = None
        return container, resolve_id

    @classonlymethod
    def __clean_path(cls, path):
        '''ensures path is Django-friendly'''
        if not path.startswith("/"):
            path = "/{}".format(path)
        if not path.endswith("/"):
            path = "{}/".format(path)
        return path

    @classonlymethod
    def get_or_create(cls, model, urlid, update=False, **field_tuples):
        try:
            logger.debug('[get_or_create] ' + str(model) + ' backlink ' + str(urlid))
            rval = model.objects.get(urlid=urlid)
            if update:
                for field in field_tuples.keys():
                    setattr(rval, field, field_tuples[field])
                rval.save()
            return rval
        except ObjectDoesNotExist:
            logger.debug('[get_or_create] creating..')
            if model is get_user_model():
                field_tuples['username'] = str(uuid.uuid4())
            return model.objects.create(urlid=urlid, is_backlink=True, **field_tuples)

    @classonlymethod
    def get_model_rdf_type(cls, model):
        if model is get_user_model():
            return "foaf:user"
        else:
            return Model.get_meta(model, "rdf_type")

    @classonlymethod
    def get_subclass_with_rdf_type(cls, type):
        '''returns Model subclass with Meta.rdf_type matching parameterised type, or None'''
        if type == 'foaf:user':
            return get_user_model()

        for subcls in Model.__subclasses__():
            if Model.get_meta(subcls, 'rdf_type') == type:
                return subcls

        return None

    @classonlymethod
    def get_permission_classes(cls, related_model, default_permissions_classes):
        '''returns the permission_classes set in the models Meta class'''
        return cls.get_meta(related_model, 'permission_classes', default_permissions_classes)

    @classonlymethod
    def get_meta(cls, model_class, meta_name, default=None):
        '''returns the models Meta class'''
        if hasattr(model_class, 'Meta'):
            meta = getattr(model_class.Meta, meta_name, default)
        else:
            meta = default
        return getattr(model_class._meta, meta_name, meta)

    @staticmethod
    def get_permissions(obj_or_model, user_or_group, filter):
        permissions = filter
        for permission_class in Model.get_permission_classes(obj_or_model, [LDPPermissions]):
            permissions = permission_class().filter_user_perms(user_or_group, obj_or_model, permissions)
        return [{'mode': {'@type': name.split('_')[0]}} for name in permissions]

    @classmethod
    def is_external(cls, value):
        '''returns True if the urlid of the value passed is from an external source'''
        try:
            return value.urlid is not None and not value.urlid.startswith(settings.SITE_URL)
        except:
            return False


class LDPSource(Model):
    federation = models.CharField(max_length=255)

    class Meta(Model.Meta):
        rdf_type = 'ldp:Container'
        ordering = ('federation',)
        container_path = 'sources'
        lookup_field = 'federation'

    def __str__(self):
        return "{}: {}".format(self.federation, self.urlid)


class Activity(Model):
    '''Models an ActivityStreams Activity'''
    aid = LDPUrlField(null=True) # activity id
    local_id = LDPUrlField()  # /inbox or /outbox full url
    payload = BinaryField()
    created_at = DateField(auto_now_add=True)

    class Meta(Model.Meta):
        container_path = "activities"
        rdf_type = 'as:Activity'

    def to_activitystream(self):
        payload = self.payload.decode("utf-8")
        data = json.loads(payload)
        return data


class Follower(Model):
    '''Models a subscription on a model. When the model is saved, an Update activity will be sent to the inbox'''
    object = models.URLField()
    inbox = models.URLField()

    def __str__(self):
        return 'Inbox ' + str(self.inbox) + ' on ' + str(self.object)

    def save(self, *args, **kwargs):
        if self.pk is None:
            logger.debug('[Follower] saving Follower ' + self.__str__())
        super(Follower, self).save(*args, **kwargs)


@receiver([post_save])
def auto_urlid(sender, instance, **kwargs):
    if isinstance(instance, Model):
        if getattr(instance, Model.slug_field(instance), None) is None:
            setattr(instance, Model.slug_field(instance), instance.pk)
            instance.save()
        if (instance.urlid is None or instance.urlid == '' or 'None' in instance.urlid):
            instance.urlid = instance.get_absolute_url()
            instance.save()


if 'djangoldp_account' not in settings.DJANGOLDP_PACKAGES:
    def webid(self):
        # an external user should have urlid set
        webid = getattr(self, 'urlid', None)
        if webid is not None and urlparse(settings.BASE_URL).netloc != urlparse(webid).netloc:
            webid = self.urlid
        # local user use user-detail URL with primary key
        else:
            webid = '{0}{1}'.format(settings.BASE_URL, reverse_lazy('user-detail', kwargs={'pk': self.pk}))
        return webid

    get_user_model().webid = webid



import threading
from urllib.parse import urlparse
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models.signals import post_save, post_delete, m2m_changed
from django.dispatch import receiver
from rest_framework.utils import model_meta

from djangoldp.models import Model, Follower

from .objects import *
from .verbs import *
import logging


logger = logging.getLogger('djangoldp')


class ActivityPubService(object):
    '''A service for sending ActivityPub notifications'''

    @classmethod
    def build_object_tree(cls, instance):
        '''builds an object tree from a parameterised instance'''
        model = type(instance)
        info = model_meta.get_field_info(model)

        if not hasattr(instance, 'urlid'):
            return

        obj = {
            "@type": Model.get_model_rdf_type(model),
            "@id": instance.urlid
        }
        if obj['@type'] is None:
            return

        # append relations
        for field_name, relation_info in info.relations.items():
            if not relation_info.to_many:
                value = getattr(instance, field_name, None)
                if value is None:
                    continue

                sub_object = {
                    "@id": value.urlid,
                    "@type": Model.get_model_rdf_type(type(value))
                }

                if sub_object['@type'] is None:
                    continue

                obj[field_name] = sub_object

        return obj

    @classmethod
    def _discover_inbox(cls, target_id):
        url = urlparse(target_id)
        return target_id.replace(url.path, "/") + "inbox/"

    @classmethod
    def send_add_activity(cls, actor, object, target):
        '''
        Sends an Add activity
        :param actor: a valid Actor object, or a user instance
        :param object: a valid ActivityStreams Object
        :param target: an object representing the target collection
        '''
        # bounds checking
        if isinstance(actor, get_user_model()):
            actor = {
                '@type': 'foaf:user',
                '@id': actor.urlid
            }

        summary = str(object['@id']) + " was added to " + str(target['@id'])

        activity = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                settings.LDP_RDF_CONTEXT
            ],
            "summary": summary,
            "type": "Add",
            "actor": actor,
            "object": object,
            "target": target
        }

        logger.debug('[Sender] sending add activity ' + str(activity))

        inbox = ActivityPubService._discover_inbox(target['@id'])

        # send request
        t = threading.Thread(target=cls.do_post, args=[inbox, activity])
        t.start()

    @classmethod
    def send_remove_activity(cls, actor, object, origin):
        '''
        Sends a Remove activity
        :param actor: a valid Actor object, or a user instance
        :param object: a valid ActivityStreams Object
        :param origin: the context the object has been removed from
        '''
        # bounds checking
        if isinstance(actor, get_user_model()):
            actor = {
                '@type': 'foaf:user',
                '@id': actor.urlid
            }

        summary = str(object['@id']) + " was removed from " + str(origin['@id'])

        activity = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                settings.LDP_RDF_CONTEXT
            ],
            "summary": summary,
            "type": "Remove",
            "actor": actor,
            "object": object,
            "origin": origin
        }

        logger.debug('[Sender] sending remove activity ' + str(activity))

        inbox = ActivityPubService._discover_inbox(origin['@id'])

        # send request
        t = threading.Thread(target=cls.do_post, args=[inbox, activity])
        t.start()

    @classmethod
    def send_create_activity(cls, actor, object, inbox):
        '''
        Sends a Create activity
        :param actor: a valid Actor object, or a user instance
        :param object: a valid ActivityStreams Object
        :param inbox: the inbox to send the activity to
        '''
        # bounds checking
        if isinstance(actor, get_user_model()):
            actor = {
                '@type': 'foaf:user',
                '@id': actor.urlid
            }

        summary = str(object['@id']) + " was created"

        activity = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                settings.LDP_RDF_CONTEXT
            ],
            "summary": summary,
            "type": "Create",
            "actor": actor,
            "object": object
        }

        logger.debug('[Sender] sending create activity ' + str(activity))

        # send request
        t = threading.Thread(target=cls.do_post, args=[inbox, activity])
        t.start()

    @classmethod
    def send_update_activity(cls, actor, object, inbox):
        '''
        Sends an Update activity
        :param actor: a valid Actor object, or a user instance
        :param object: a valid ActivityStreams Object
        :param inbox: the inbox to send the activity to
        '''
        # bounds checking
        if isinstance(actor, get_user_model()):
            actor = {
                '@type': 'foaf:user',
                '@id': actor.urlid
            }

        summary = str(object['@id']) + " was updated"

        activity = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                settings.LDP_RDF_CONTEXT
            ],
            "summary": summary,
            "type": "Update",
            "actor": actor,
            "object": object
        }

        logger.debug('[Sender] sending update activity ' + str(activity))

        # send request
        t = threading.Thread(target=cls.do_post, args=[inbox, activity])
        t.start()

    @classmethod
    def send_delete_activity(cls, actor, object, inbox):
        '''
        Sends a Remove activity
        :param actor: a valid Actor object, or a user instance
        :param object: a valid ActivityStreams Object
        :param inbox: the inbox to send the activity to
        '''
        # bounds checking
        if isinstance(actor, get_user_model()):
            actor = {
                '@type': 'foaf:user',
                '@id': actor.urlid
            }

        summary = str(object['@id']) + " was deleted"

        activity = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                settings.LDP_RDF_CONTEXT
            ],
            "summary": summary,
            "type": "Delete",
            "actor": actor,
            "object": object
        }

        logger.debug('[Sender] sending delete activity ' + str(activity))

        # send request
        t = threading.Thread(target=cls.do_post, args=[inbox, activity])
        t.start()

    @classmethod
    def do_post(cls, url, activity, auth=None):
        '''makes a POST request to passed url'''
        headers = {'Content-Type': 'application/ld+json'}
        response = None
        try:
            response = requests.post(url, data=json.dumps(activity), headers=headers)
            logger.debug('[Sender] sent, receiver responded ' + response.text)
        except:
            logger.error('Failed to deliver backlink to ' + str(url) +', was attempting ' + str(activity))
        return response


def _check_instance_for_backlinks(sender, instance):
    '''Auxiliary function returns a set of backlink targets from paramertised instance'''
    info = model_meta.get_field_info(sender)

    # bounds checking
    if not hasattr(instance, 'urlid') or Model.get_model_rdf_type(sender) is None:
        return {}

    # check each foreign key for a distant resource
    targets = set()
    for field_name, relation_info in info.relations.items():
        if not relation_info.to_many:
            value = getattr(instance, field_name, None)
            logger.debug('[Sender] model has relation ' + str(value))
            if value is not None and Model.is_external(value):
                target_type = Model.get_model_rdf_type(type(value))

                if target_type is None:
                    continue

                targets.add(ActivityPubService._discover_inbox(value.urlid))

    # append Followers as targets
    followers = Follower.objects.filter(object=instance.urlid)
    for follower in followers:
        targets.add(follower.inbox)

    logger.debug('[Sender] built set of targets: ' + str(targets))
    return targets


@receiver([post_save])
def check_save_for_backlinks(sender, instance, created, **kwargs):
    if getattr(settings, 'SEND_BACKLINKS', True) and getattr(instance, 'allow_create_backlink', False) \
            and not Model.is_external(instance) \
            and getattr(instance, 'username', None) != 'hubl-workaround-493':
        logger.debug("[Sender] Received created non-backlink instance " + str(instance) + "(" + str(sender) + ")")
        targets = _check_instance_for_backlinks(sender, instance)

        if len(targets) > 0:
            obj = ActivityPubService.build_object_tree(instance)
            actor = {
                "type": "Service",
                "name": "Backlinks Service"
            }
            # Create Activity
            if created:
                for target in targets:
                    ActivityPubService.send_create_activity(actor, obj, target)
                    Follower.objects.create(object=obj['@id'], inbox=target)
            # Update Activity
            else:
                for target in targets:
                    ActivityPubService.send_update_activity(actor, obj, target)
                    if not Follower.objects.filter(object=obj['@id'], inbox=target).exists():
                        Follower.objects.create(object=obj['@id'], inbox=target)


@receiver([post_delete])
def check_delete_for_backlinks(sender, instance, **kwargs):
    if getattr(settings, 'SEND_BACKLINKS', True) and getattr(instance, 'allow_create_backlink', False) \
            and getattr(instance, 'username', None) != 'hubl-workaround-493':
        logger.debug("[Sender] Received deleted non-backlink instance " + str(instance) + "(" + str(sender) + ")")
        targets = _check_instance_for_backlinks(sender, instance)

        if len(targets) > 0:
            for target in targets:
                ActivityPubService.send_delete_activity({
                    "type": "Service",
                    "name": "Backlinks Service"
                }, {
                    "@id": instance.urlid,
                    "@type": Model.get_model_rdf_type(sender)
                }, target)

    # remove any Followers on this resource
    urlid = getattr(instance, 'urlid', None)
    if urlid is not None:
        for follower in Follower.objects.filter(object=urlid):
            follower.delete()


@receiver([m2m_changed])
def check_m2m_for_backlinks(sender, instance, action, *args, **kwargs):
    def build_targets(query_set):
        '''analyses parameterised queryset (removed or added members) for backlinks'''
        targets = []

        for obj in query_set:
            condition = Model.is_external(obj) and getattr(obj, 'allow_create_backlink', False)
            if action == "post_add":
                condition = condition and not Model.is_external(instance)

            if condition:
                targets.append({
                    "@type": member_rdf_type,
                    "@id": obj.urlid
                })

        return targets

    if getattr(settings, 'SEND_BACKLINKS', True):
        member_model = kwargs['model']
        pk_set = kwargs['pk_set']

        # we can only send backlinks on pre_clear because on post_clear the objects are gone
        if action != "pre_clear" and pk_set is None:
            return
        member_rdf_type = Model.get_model_rdf_type(member_model)
        container_rdf_type = Model.get_model_rdf_type(type(instance))

        if member_rdf_type is None:
            return
        if container_rdf_type is None:
            return

        # build list of targets (models affected by the change)
        if action == "pre_clear":
            pk_set = sender.objects.all().values_list(member_model.__name__.lower(), flat=True)
        query_set = member_model.objects.filter(pk__in=pk_set)
        targets = build_targets(query_set)

        logger.debug('[Sender] checking many2many for backlinks')
        logger.debug('[Sender] built targets: ' + str(targets))

        if len(targets) > 0:
            obj = {
                "@type": container_rdf_type,
                "@id": instance.urlid
            }
            if action == 'post_add':
                for target in targets:
                    ActivityPubService.send_add_activity({
                        "type": "Service",
                        "name": "Backlinks Service"
                    }, obj, target)

            elif action == "post_remove" or action == "pre_clear":
                for target in targets:
                    ActivityPubService.send_remove_activity({
                        "type": "Service",
                        "name": "Backlinks Service"
                    }, obj, target)



import json
from django.contrib.auth import get_user_model
from rest_framework.test import APIClient, APITestCase
from djangoldp.tests.models import Circle, CircleMember, Project, UserProfile
from djangoldp.models import Activity, Follower


class TestsInbox(APITestCase):

    def setUp(self):
        self.client = APIClient(enforce_csrf_checks=True)

    #
    #   CREATE ACTIVITY
    #
    def test_create_activity_circle(self):
        # a local user has been set as the owner of a distant circle
        user = get_user_model().objects.create(username='john', email='jlennon@beatles.com', password='glass onion')
        UserProfile.objects.create(user=user)

        payload = {
          "@context": [
              "https://www.w3.org/ns/activitystreams",
              {"hd": "http://happy-dev.fr/owl/#"}
          ],
          "summary": "A circle was created",
          "type": "Create",
          "actor": {
            "type": "Service",
            "name": "Backlinks Service"
          },
          "object": {
            "@type": "hd:circle",
            "@id": "https://distant.com/circles/1/",
            "owner": {
                "@type": "foaf:user",
                "@id": user.urlid
            }
          }
        }

        response = self.client.post('/inbox/',
                                    data=json.dumps(payload), content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)

        # assert that the circle was created and the user associated as owner
        circles = Circle.objects.all()
        activities = Activity.objects.all()
        self.assertEquals(len(circles), 1)
        self.assertEquals(len(activities), 1)
        self.assertIn("https://distant.com/circles/1/", circles.values_list('urlid', flat=True))
        self.assertEqual(circles[0].owner, user)
        self.assertIn(response["Location"], activities.values_list('urlid', flat=True))

    #
    #   ADD ACTIVITIES
    #
    # project model has a direct many-to-many with User
    def test_add_activity_project(self):
        # a local user has joined a distant project
        user = get_user_model().objects.create(username='john', email='jlennon@beatles.com', password='glass onion')
        UserProfile.objects.create(user=user)

        payload = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                {"hd": "http://happy-dev.fr/owl/#"}
            ],
            "summary": user.get_full_name() + " was added to Test Project",
            "type": "Add",
            "actor": {
              "type": "Service",
              "name": "Backlinks Service"
            },
            "object": {
                "@type": "hd:project",
                "@id": "https://distant.com/projects/1/"
            },
            "target": {
                "@type": "foaf:user",
                "name": user.get_full_name(),
                "@id": user.urlid
            }
        }

        response = self.client.post('/inbox/',
                                    data=json.dumps(payload), content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)

        # assert that the project backlink(s) & activity were created
        projects = Project.objects.all()
        user_projects = user.projects.all()
        activities = Activity.objects.all()
        self.assertEquals(len(projects), 1)
        self.assertEquals(len(user_projects), 1)
        self.assertEquals(len(activities), 1)
        self.assertIn("https://distant.com/projects/1/", projects.values_list('urlid', flat=True))
        self.assertIn("https://distant.com/projects/1/", user_projects.values_list('urlid', flat=True))
        self.assertIn(response["Location"], activities.values_list('urlid', flat=True))

    # circle model has a many-to-many with user, through an intermediate model
    def test_add_activity_circle(self):
        # a local user has joined a distant circle
        user = get_user_model().objects.create(username='john', email='jlennon@beatles.com', password='glass onion')
        UserProfile.objects.create(user=user)

        payload = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                {"hd": "http://happy-dev.fr/owl/#"}
            ],
            "summary": user.get_full_name() + " was added to Test Circle",
            "type": "Add",
            "actor": {
              "type": "Service",
              "name": "Backlinks Service"
            },
            "object": {
                "@type": "hd:circlemember",
                "@id": "https://distant.com/circle-members/1/",
                "user": {
                  "@type": "foaf:user",
                  "@id": user.urlid
                },
                "circle": {
                    "@type": "hd:circle",
                    "@id": "https://distant.com/circles/1/"
                }
            },
            "target": {
                "@type": "foaf:user",
                "name": user.get_full_name(),
                "@id": user.urlid
            }
        }

        response = self.client.post('/inbox/',
                                    data=json.dumps(payload), content_type='application/ld+json;profile="https://www.w3.org/ns/activitystreams"')
        self.assertEqual(response.status_code, 201)

        # assert that the circle backlink(s) & activity were created
        circles = Circle.objects.all()
        user_circles = user.circles.all()
        activities = Activity.objects.all()
        self.assertEquals(len(circles), 1)
        self.assertEquals(len(user_circles), 1)
        self.assertEquals(len(activities), 1)
        self.assertIn("https://distant.com/circles/1/", circles.values_list('urlid', flat=True))
        self.assertIn("https://distant.com/circle-members/1/", user_circles.values_list('urlid', flat=True))
        self.assertIn(response["Location"], activities.values_list('urlid', flat=True))

    # test sending an add activity when the backlink already exists
    def test_add_activity_object_already_added(self):
        # a local user has joined a distant circle
        user = get_user_model().objects.create(username='john', email='jlennon@beatles.com', password='glass onion')
        UserProfile.objects.create(user=user)

        # ..but the receiver already knows about it
        circle = Circle.objects.create(urlid="https://distant.com/circles/1/")
        CircleMember.objects.create(urlid="https://distant.com/circle-members/1/", circle=circle, user=user)

        payload = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                {"hd": "http://happy-dev.fr/owl/#"}
            ],
            "summary": user.get_full_name() + " was added to Test Circle",
            "type": "Add",
            "actor": {
                "type": "Service",
                "name": "Backlinks Service"
            },
            "object": {
                "@type": "hd:circlemember",
                "@id": "https://distant.com/circle-members/1/",
                "user": {
                    "@type": "foaf:user",
                    "@id": user.urlid
                },
                "circle": {
                    "@type": "hd:circle",
                    "@id": "https://distant.com/circles/1/"
                }
            },
            "target": {
                "@type": "foaf:user",
                "name": user.get_full_name(),
                "@id": user.urlid
            }
        }

        response = self.client.post('/inbox/',
                                    data=json.dumps(payload),
                                    content_type='application/ld+json;profile="https://www.w3.org/ns/activitystreams"')
        self.assertEqual(response.status_code, 201)

        # assert that the circle backlink(s) & activity were created
        circles = Circle.objects.all()
        user_circles = user.circles.all()
        activities = Activity.objects.all()
        self.assertEquals(len(circles), 1)
        self.assertEquals(len(user_circles), 1)
        self.assertEquals(len(activities), 1)
        self.assertIn("https://distant.com/circles/1/", circles.values_list('urlid', flat=True))
        self.assertIn("https://distant.com/circle-members/1/", user_circles.values_list('urlid', flat=True))
        self.assertIn(response["Location"], activities.values_list('urlid', flat=True))

    # TODO: adding to a model which has multiple relationships with this RDF type

    # error behaviour - unknown model
    def test_add_activity_unknown(self):
        # a local user has joined a distant circle
        user = get_user_model().objects.create(username='john', email='jlennon@beatles.com', password='glass onion')
        UserProfile.objects.create(user=user)

        payload = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                {"hd": "http://happy-dev.fr/owl/#"}
            ],
            "summary": user.get_full_name() + " was added to Test Circle",
            "type": "Add",
            "actor": {
              "type": "Service",
              "name": "Backlinks Service"
            },
            "object": {
                "@type": "hd:somethingunknown",
                "@id": "https://distant.com/somethingunknown/1/"
            },
            "target": {
                "@type": "foaf:user",
                "name": user.get_full_name(),
                "@id": user.urlid
            }
        }

        response = self.client.post('/inbox/',
                                    data=json.dumps(payload), content_type='application/ld+json;profile="https://www.w3.org/ns/activitystreams"')
        self.assertEqual(response.status_code, 404)

    #
    #   REMOVE & DELETE ACTIVITIES
    #
    # project model has a direct many-to-many with User
    def test_remove_activity_project_using_origin(self):
        # a local user has a distant project attached
        user = get_user_model().objects.create(username='john', email='jlennon@beatles.com', password='glass onion')
        UserProfile.objects.create(user=user)
        project = Project.objects.create(urlid="https://distant.com/projects/1/")
        user.projects.add(project)

        payload = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                {"hd": "http://happy-dev.fr/owl/#"}
            ],
            "summary": user.get_full_name() + " removed Test Project",
            "type": "Remove",
            "actor": {
              "type": "Service",
              "name": "Backlinks Service"
            },
            "object": {
                "@type": "hd:project",
                "@id": "https://distant.com/projects/1/"
            },
            "origin": {
                "@type": "foaf:user",
                "name": user.get_full_name(),
                "@id": user.urlid
            }
        }

        response = self.client.post('/inbox/',
                                    data=json.dumps(payload),
                                    content_type='application/ld+json;profile="https://www.w3.org/ns/activitystreams"')
        self.assertEqual(response.status_code, 201)

        # assert that the circle backlink(s) were removed & activity were created
        projects = Project.objects.all()
        user_projects = user.projects.all()
        activities = Activity.objects.all()
        self.assertEquals(len(projects), 1)
        self.assertEquals(len(user_projects), 0)
        self.assertEquals(len(activities), 1)
        self.assertIn("https://distant.com/projects/1/", projects.values_list('urlid', flat=True))
        self.assertIn(response["Location"], activities.values_list('urlid', flat=True))

    # TODO: test_remove_activity_project_using_target
    # TODO: error behaviour - project does not exist on user

    # Delete CircleMember
    def test_delete_activity_circle_using_origin(self):
        # a local user has a distant circle attached
        user = get_user_model().objects.create(username='john', email='jlennon@beatles.com', password='glass onion')
        UserProfile.objects.create(user=user)
        circle = Circle.objects.create(urlid="https://distant.com/circles/1/", allow_create_backlink=False)
        CircleMember.objects.create(urlid="https://distant.com/circle-members/1/",circle=circle, user=user)

        payload = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                {"hd": "http://happy-dev.fr/owl/#"}
            ],
            "summary": "CircleMember was deleted",
            "type": "Delete",
            "actor": {
              "type": "Service",
              "name": "Backlinks Service"
            },
            "object": {
                "@type": "hd:circlemember",
                "@id": "https://distant.com/circle-members/1/",
                "user": {
                    "@type": "foaf:user",
                    "@id": user.urlid
                },
                "circle": {
                    "@type": "hd:circle",
                    "@id": "https://distant.com/circles/1/"
                }
            }
        }

        response = self.client.post('/inbox/',
                                    data=json.dumps(payload),
                                    content_type='application/ld+json;profile="https://www.w3.org/ns/activitystreams"')
        self.assertEqual(response.status_code, 201)

        # assert that the CircleMember was deleted and activity was created
        circles = Circle.objects.all()
        user_circles = user.circles.all()
        activities = Activity.objects.all()
        self.assertEquals(len(circles), 1)
        self.assertEquals(len(user_circles), 0)
        self.assertEquals(len(activities), 1)
        self.assertIn("https://distant.com/circles/1/", circles.values_list('urlid', flat=True))
        self.assertIn(response["Location"], activities.values_list('urlid', flat=True))

    # TODO: test_delete_activity_circle_using_target

    #
    #   UPDATE Activities
    #
    def test_update_activity_circle(self):
        # a local user was set as the owner of a distant circle, but the owner has been changed
        user = get_user_model().objects.create(username='john', email='jlennon@beatles.com', password='glass onion')
        UserProfile.objects.create(user=user)

        circle = Circle.objects.create(urlid="https://distant.com/circles/1/", owner=user)
        self.assertEqual(circle.owner, user)

        payload = {
          "@context": [
              "https://www.w3.org/ns/activitystreams",
              {"hd": "http://happy-dev.fr/owl/#"}
          ],
          "summary": "A circle was updated",
          "type": "Update",
          "actor": {
            "type": "Service",
            "name": "Backlinks Service"
          },
          "object": {
            "@type": "hd:circle",
            "@id": "https://distant.com/circles/1/",
            "owner": {
                "@type": "foaf:user",
                "@id": "https://distant.com/users/1/"
            }
          }
        }

        response = self.client.post('/inbox/',
                                    data=json.dumps(payload), content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)

        # assert that the circle was created and the user associated as owner
        circles = Circle.objects.all()
        activities = Activity.objects.all()
        users = get_user_model().objects.all()
        self.assertEquals(len(circles), 1)
        self.assertEquals(len(activities), 1)
        self.assertEquals(len(users), 2)
        distant_user = get_user_model().objects.get(urlid="https://distant.com/users/1/")
        self.assertIn("https://distant.com/circles/1/", circles.values_list('urlid', flat=True))
        self.assertEqual(circles[0].owner, distant_user)
        self.assertIn(response["Location"], activities.values_list('urlid', flat=True))

    #
    #   FOLLOW activities
    #
    def test_follow_activity(self):
        # a local user was set as the owner of a distant circle, but the owner has been changed
        user = get_user_model().objects.create(username='john', email='jlennon@beatles.com', password='glass onion')
        UserProfile.objects.create(user=user)

        circle = Circle.objects.create(description='Test Description')

        payload = {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                {"hd": "http://happy-dev.fr/owl/#"}
            ],
            "summary": user.urlid + " followed " + circle.urlid,
            "type": "Follow",
            "actor": {
                "type": "Service",
                "name": "Backlinks Service",
                "inbox": "http://127.0.0.1:8000/inbox/"
            },
            "object": {
                "@type": "hd:circle",
                "@id": circle.urlid
            }
        }

        response = self.client.post('/inbox/',
                                    data=json.dumps(payload), content_type='application/ld+json')
        self.assertEqual(response.status_code, 201)

        # assert that Follower was created with correct values
        followers = Follower.objects.all()
        activities = Activity.objects.all()
        self.assertEquals(len(followers), 1)
        self.assertEquals(len(activities), 1)
        self.assertIn(response["Location"], activities.values_list('urlid', flat=True))
        follower = followers[0]
        self.assertEqual("http://127.0.0.1:8000/inbox/", follower.inbox)
        self.assertEqual(circle.urlid, follower.object)

    # test Followers are auto-deleted when the object they're following is deleted
    def test_follower_auto_delete(self):
        user = get_user_model().objects.create(username='john', email='jlennon@beatles.com', password='glass onion')
        UserProfile.objects.create(user=user)

        circle = Circle.objects.create(description='Test Description')
        Follower.objects.create(object=circle.urlid, inbox="http://127.0.0.1:8000/inbox/")
        followers = Follower.objects.all()
        self.assertEquals(len(followers), 1)
        circle.delete()
        followers = Follower.objects.all()
        self.assertEquals(len(followers), 0)

    #
    #   GET Inbox
    #
    def test_get_inbox(self):
        response = self.client.get('/inbox/')
        self.assertEqual(response.status_code, 405)



import unittest

from django.test import TestCase

from djangoldp.models import Model
from djangoldp.tests.models import Dummy, LDPDummy


class LDPModelTest(TestCase):

    def test_class_not_inheriting_ldp_model(self):
        dummy = Dummy.objects.create(some="text")
        self.assertEquals("/dummys/", Model.container_id(dummy))
        self.assertEquals("/dummys/{}/".format(dummy.slug), Model.resource_id(dummy))

    def test_class_inheriting_ldp_model(self):
        dummy = LDPDummy.objects.create(some="text")
        self.assertEquals("/ldpdummys/", dummy.get_container_id())
        self.assertEquals("http://happy-dev.fr/ldpdummys/{}/".format(dummy.pk), dummy.get_absolute_url())
        self.assertEquals("/ldpdummys/", Model.container_id(dummy))
        self.assertEquals("/ldpdummys/{}/".format(dummy.pk), Model.resource_id(dummy))

    def test_from_resolve_id(self):
        saved_instance = Dummy.objects.create(some="text", slug="someid")
        result = Model.resolve_id("/dummys/{}/".format(saved_instance.slug))
        self.assertEquals(saved_instance, result)

    def test_resolve_container(self):
        result = Model.resolve_container("/dummys/")
        self.assertEquals(Dummy, result)

    def test_auto_url(self):
        from django.urls import get_resolver
        dummy = LDPDummy.objects.create(some="text")
        view_name = '{}-list'.format(dummy._meta.object_name.lower())
        path = 'http://happy-dev.fr/{}{}/'.format(get_resolver().reverse_dict[view_name][0][0][0], dummy.pk)

        self.assertEquals(path, dummy.get_absolute_url())
